#include <inttypes.h>
#include <math.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "`fwrite_le`"

// constants
#if defined(_WIN32)
# define __ANSI_CLEAR_STRING "\r"
#elif defined(__unix__) || defined(__linux__)
# define __ANSI_CLEAR_STRING "\x1B[2K\r"
#else
# define __ANSI_CLEAR_STRING "\n"
#endif
const char* ANSI_CLEAR = __ANSI_CLEAR_STRING;

#define OUTPUT_FILE `output_file`

#define SAMPLE_RATE `sample_rate`
#define ORIGINAL_SAMPLE_RATE `original_sample_rate`
#define BIT_DEPTH `bit_depth`
#define IS_SIGNED `is_signed`
#define CHANNELS `channels`
#define LENGTH `length`

#if BIT_DEPTH <= 8
# define SAMPLE_TYPE uint8_t
#elif BIT_DEPTH >= 16
# if IS_SIGNED
#  define SAMPLE_TYPE int16_t
# else
#  define SAMPLE_TYPE uint16_t
# endif
#endif

#define PRODUCT `wav_product`
#define FREQUENCY_OF_STATUS_REPORTING 5000

#define SEQUENTIAL_MODE `sequential_mode`
#define BLOCK_SIZE_BYTES `block_size`

#define FP_RETURN_TYPE `fp_return_type`
#define PRECALCULATED_RATIO `precalculated_ratio`

#define BIT_DEPTH_LIMITER     ((1 << BIT_DEPTH) - 1)
#define PCM_COEFFICIENT       ((1 << (BIT_DEPTH - 1)) - 1)
#define unsigned_to_signed(x) (x - PCM_COEFFICIENT)
#define signed_to_unsigned(x) (x + PCM_COEFFICIENT)

// macros
#define ALLOCATE_MEMORY(nmemb)                                                \
    SAMPLE_TYPE* buffer = malloc(                                             \
        (size_t) nmemb * sizeof(SAMPLE_TYPE)                                  \
    );                                                                        \
                                                                              \
    if (buffer == NULL) {                                                     \
        perror("malloc");                                                     \
        return EXIT_FAILURE;                                                  \
    }

// global variables
#define SILENT_MODE `silent_mode`
#define VERBOSE_MODE `verbose_mode`

// function prototypes
#if FP_RETURN_TYPE
long double
#else
SAMPLE_TYPE
#endif
bytebeat(long double time);

// function implementations
#if FP_RETURN_TYPE
long double
#else
SAMPLE_TYPE
#endif
bytebeat(long double time)
{
#if PRECALCULATED_RATIO
	`final_sample_rate_code`
#elif ORIGINAL_SAMPLE_RATE != SAMPLE_RATE
# if SAMPLE_RATE > ORIGINAL_SAMPLE_RATE
	time /= ((long double) SAMPLE_RATE) / ((long double) ORIGINAL_SAMPLE_RATE);
# else
	time *= ((long double) ORIGINAL_SAMPLE_RATE / ((long double) SAMPLE_RATE));
# endif
#endif
	uintmax_t t = (uintmax_t) time;
	`bytebeat_contents`;
}

int
main(void)
{
	// * log -> welcome
#if !SILENT_MODE
	printf(":: C bytebeat generator runtime unit\n");
	fflush(stdout);
#endif

#if !SILENT_MODE
	const uintmax_t seconds = LENGTH / SAMPLE_RATE,
		samples = LENGTH % SAMPLE_RATE;

	printf(
	  "\n"
	  "Sample rate: %d Hz\n"
	  "Channels: %d%s\n"
	  "Bit depth: %ssigned %d-bit\n"
	  "Duration: ",
	  SAMPLE_RATE,
	  CHANNELS,
	  CHANNELS == 1 ? " (mono)" : (CHANNELS > 2 ? "" : " (stereo)"),
	  IS_SIGNED ? "" : "un",
	  BIT_DEPTH);

	if (seconds > 0)
		if (seconds >= 3600)
			printf(
			  "%" PRIuMAX ":%02" PRIuMAX ":%02" PRIuMAX,
			  seconds / 3600,
			  (seconds / 60) % 60,
			  seconds % 60);
		else if (seconds >= 60)
			printf("%" PRIuMAX ":%02" PRIuMAX, seconds / 60, seconds % 60);
		else
			printf("%" PRIuMAX " seconds", seconds);

	if (seconds > 0 && samples > 0)
		printf(" + ");

	if (samples > 0)
		printf("%" PRIuMAX " samples", samples);

	printf(
#if VERBOSE_MODE || SEQUENTIAL_MODE
		"\n"
#endif
		"\n"
	);

	fflush(stdout);
#endif

	// * write WAVE headers
	// 0. log
#if SEQUENTIAL_MODE
	printf("Writing WAVE headers...\n");
#endif

	// 1. open file
	FILE* output_file = fopen(OUTPUT_FILE, "wb");
	if (output_file == NULL) {
		fflush(stdout);
		perror("fopen");
		return 1;
	}

	// 2. prepare variables
	uint32_t buffer_size = PRODUCT,
		file_length =
			4 * 4   /* 4 strings of 4 characters */ +
			5 * 4   /* 4 uint32_t values */         +
			4 * 2   /* 5 uint16_t values */         +
			PRODUCT /* sample data */,
		fmt_data_length = 16 /* <--
		 * length of format data before this value
		 * in the file format structure
		 */,
		sample_rate = SAMPLE_RATE,
		byte_rate = (SAMPLE_RATE * BIT_DEPTH * CHANNELS) / 8;
	uint16_t fmt_type = 1, // format type is PCM
		channels = CHANNELS,
		block_align = (BIT_DEPTH * CHANNELS) / 8,
		bit_depth = BIT_DEPTH > 8 ? BIT_DEPTH : 8;

	// 3. write headers
	// <L = Little-endian or B = Big-endian> : <name> : <bytes of field>
	fwrite   ("RIFF",           1, 4, output_file); // B : ChunkID       : 4
	fwrite_le(&file_length,     4, 1, output_file); // L : ChunkSize     : 4
	fwrite   ("WAVE",           1, 4, output_file); // B : Format        : 4
	fwrite   ("fmt ",           1, 4, output_file); // B : Subchunk1ID   : 4
	fwrite_le(&fmt_data_length, 4, 1, output_file); // L : Subchunk1Size : 4
	fwrite_le(&fmt_type,        2, 1, output_file); // L : AudioFormat   : 2
	fwrite_le(&channels,        2, 1, output_file); // L : NumChannels   : 2
	fwrite_le(&sample_rate,     4, 1, output_file); // L : SampleRate    : 4
	fwrite_le(&byte_rate,       4, 1, output_file); // L : ByteRate      : 4
	fwrite_le(&block_align,     2, 1, output_file); // L : BlockAlign    : 2
	fwrite_le(&bit_depth,       2, 1, output_file); // L : BitsPerSample : 2
	fwrite   ("data",           1, 4, output_file); // B : Subchunk2ID   : 4
	fwrite_le(&buffer_size,     4, 1, output_file); // L : Subchunk2Size : 4

	// 4. write sample data
#if SEQUENTIAL_MODE
	size_t BLOCK_SIZE = BLOCK_SIZE_BYTES / (sizeof(SAMPLE_TYPE) /
		sizeof(uint8_t));
	if (BLOCK_SIZE < 1) {
		printf("The block size %" PRIuMAX " is too small, should be at least "
		       "%" PRIuMAX " bytes\n", (uintmax_t) BLOCK_SIZE_BYTES,
		       (uintmax_t) ((sizeof(SAMPLE_TYPE) + sizeof(uint8_t) - 1) /
		                     sizeof(uint8_t)));
		return EXIT_FAILURE;
	}

	const size_t MAX = (PRODUCT + (BLOCK_SIZE - 1)) / BLOCK_SIZE;
#endif

    // * allocate heap for sample data
#if SEQUENTIAL_MODE
	size_t calc_block_size = BLOCK_SIZE;
	ALLOCATE_MEMORY(calc_block_size)
#else
	ALLOCATE_MEMORY(PRODUCT)
#endif

#if SEQUENTIAL_MODE
	size_t time = 0;
	for (size_t seq = 0; seq < MAX; seq++) {
		if ((time + BLOCK_SIZE) >= PRODUCT)
			calc_block_size = PRODUCT - time;
#endif

		// * bytebeat generating loop
		const uintmax_t product_minus_1 = PRODUCT - 1,
			bit_depth_limiter = BIT_DEPTH_LIMITER
#if FP_RETURN_TYPE
				+ 1
#endif
#if BIT_DEPTH < 8
			,
			bit_depth_stretch = ((long double) BIT_DEPTH) / 8.L
#endif
			;

#if SEQUENTIAL_MODE
		for (size_t idx = 0; idx < BLOCK_SIZE && time < PRODUCT; idx++,
		     time++) {
#else
		for (size_t time = 0; time < PRODUCT; time++) {
#endif
			// 1. generate audio data
#if FP_RETURN_TYPE
			long double bytebeat_res =
				floor(bytebeat(floor((long double) time)));
#elif IS_SIGNED
			intmax_t bytebeat_res =
				(intmax_t) bytebeat(floor((long double) time));
#else
			uintmax_t bytebeat_res =
				(uintmax_t) bytebeat(floor((long double) time));
#endif

			// 2. if signed, then wrap up to unsigned
#if IS_SIGNED
			bytebeat_res = signed_to_unsigned(bytebeat_res);
#endif

			// 3. convert audio data to sample
			SAMPLE_TYPE sample_res = (SAMPLE_TYPE)
#if FP_RETURN_TYPE
				fmodl(bytebeat_res, bit_depth_limiter);
#else
				((uintmax_t) bytebeat_res & bit_depth_limiter);
#endif

			// 4. if bit depth is less than 8, stretch it
#if BIT_DEPTH < 8
			sample_res = (SAMPLE_TYPE)
				((long double) sample_res * bit_depth_stretch);
#endif

			// 5. save sample into buffer
#if SEQUENTIAL_MODE
			buffer[idx] = sample_res;
#else
			buffer[time] = sample_res;
#endif

			// 6. log
#if VERBOSE_MODE
			if (time % FREQUENCY_OF_STATUS_REPORTING == 0 ||
				time >= product_minus_1 /* or if writing last sample */) {
				printf(
				  "%sremaining samples = %18" PRIuMAX " (%3.2Lf%% done)"
#if SEQUENTIAL_MODE
				  " (part %" PRIuMAX "/%" PRIuMAX ")"
#endif
				  ,
				  ANSI_CLEAR,
				  product_minus_1 - time,
				  ((long double) time * 100) / (long double) PRODUCT
#if SEQUENTIAL_MODE
				  , (uintmax_t) seq + 1, (uintmax_t) MAX
#endif
				  );
				fflush(stdout);
			}
#endif
		}

#if !SILENT_MODE
		printf("%s", ANSI_CLEAR);

		// 5. log
#if !(SEQUENTIAL_MODE && VERBOSE_MODE)
#if SEQUENTIAL_MODE
		if (seq == 0)
#endif
			printf(
#if !SEQUENTIAL_MODE
				"\n"
#endif
				"Writing out file " OUTPUT_FILE "...\n");
#endif
		fflush(stdout);
#endif

		// * save the sample data into the file
		fwrite_le(buffer, sizeof(SAMPLE_TYPE),
#if SEQUENTIAL_MODE
			calc_block_size,
#else
			PRODUCT,
#endif
			output_file);
#if SEQUENTIAL_MODE
		fflush(output_file);
#endif

#if SEQUENTIAL_MODE
	}
#endif

	// * free allocated heap
	free(buffer);

	// 6. close file
	fclose(output_file);

	// * end of program
#if !SILENT_MODE
	printf(
#if SEQUENTIAL_MODE && VERBOSE_MODE
		"\n"
#endif
		"Done!\n");
#endif

	return EXIT_SUCCESS;
}
